@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# © Copyright EnterpriseDB UK Limited 2011-2023
+# © Copyright EnterpriseDB UK Limited 2011-2025
 #
 # This file is part of Barman.
 #
@@ -25,6 +25,7 @@ import logging
 import os
 import shutil
 import tempfile
+from collections import defaultdict
 from contextlib import closing
 from glob import glob
 
@@ -65,7 +66,11 @@ from barman.utils import (
     force_str,
     fsync_dir,
     fsync_file,
+    get_backup_id_from_target_lsn,
+    get_backup_id_from_target_time,
+    get_backup_id_from_target_tli,
     get_backup_info_from_name,
+    get_last_backup_id,
     human_readable_timedelta,
     pretty_size,
 )
@@ -368,12 +373,9 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
             default to :attr:`DEFAULT_STATUS_FILTER`.
         :return str|None: ID of the backup
         """
-        available_backups = self.get_available_backups(status_filter)
-        if len(available_backups) == 0:
-            return None
-
-        ids = sorted(available_backups.keys())
-        return ids[-1]
+        available_backups = self.get_available_backups(status_filter).values()
+        backup_id = get_last_backup_id(available_backups)
+        return backup_id
 
     def get_last_full_backup_id(self, status_filter=DEFAULT_STATUS_FILTER):
         """
@@ -385,7 +387,7 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
         """
         available_full_backups = list(
             filter(
-                lambda backup: backup.is_full_and_eligible_for_incremental(),
+                lambda backup: backup.is_full,
                 self.get_available_backups(status_filter).values(),
             )
         )
@@ -427,6 +429,61 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
         if backup_info is not None:
             return backup_info.backup_id
 
+    def get_closest_backup_id_from_target_time(
+        self, target_time, target_tli, status_filter=DEFAULT_STATUS_FILTER
+    ):
+        """
+        Get the id of a backup according to the time passed as the recovery target
+        *target_time*, and in the given *target_tli*, if specified.
+
+        :param str target_time: The target value with timestamp format
+            ``%Y-%m-%d %H:%M:%S`` with or without timezone.
+        :param int|None target_tli: The target timeline, if a specific one is required.
+        :param tuple[str, ...] status_filter: The status of the backup to return.
+        :return str|None: ID of the backup.
+        """
+
+        available_backups = self.get_available_backups(status_filter).values()
+        backup_id = get_backup_id_from_target_time(
+            available_backups, target_time, target_tli
+        )
+        return backup_id
+
+    def get_closest_backup_id_from_target_lsn(
+        self, target_lsn, target_tli, status_filter=DEFAULT_STATUS_FILTER
+    ):
+        """
+        Get the id of a backup according to the lsn passed as the recovery target
+        *target_lsn*, and in the given *target_tli*, if specified.
+
+        :param str target_lsn: The target value with lsn format, e.g.,
+            ``3/64000000``.
+        :param int|None target_tli: The target timeline, if a specific one is required.
+        :param tuple[str, ...] status_filter: The status of the backup to return.
+        :return str|None: ID of the backup.
+        """
+        available_backups = self.get_available_backups(status_filter).values()
+        backup_id = get_backup_id_from_target_lsn(
+            available_backups, target_lsn, target_tli
+        )
+        return backup_id
+
+    def get_last_backup_id_from_target_tli(
+        self, target_tli, status_filter=DEFAULT_STATUS_FILTER
+    ):
+        """
+        Get the id of a backup according to the timeline passed as the recovery target
+        *target_tli*.
+
+        :param int target_tli: The target timeline.
+        :param tuple[str, ...] status_filter: The status of the backup to return.
+        :return str|None: ID of the backup.
+        """
+
+        available_backups = self.get_available_backups(status_filter).values()
+        backup_id = get_backup_id_from_target_tli(available_backups, target_tli)
+        return backup_id
+
     def put_delete_annotation(self, backup_id):
         """
         Add a delete annotation to the specified backup.
@@ -735,9 +792,9 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
         if parent_backup_info:
             if parent_backup_info.summarize_wal != "on":
                 raise BackupException(
-                    "The specified backup is not eligible as a parent for an "
+                    "Backup ID %s is not eligible as a parent for an "
                     "incremental backup because WAL summaries were not enabled "
-                    "when that backup was taken."
+                    "when that backup was taken." % parent_backup_info.backup_id
                 )
             if parent_backup_info.compression is not None:
                 raise BackupException(
@@ -932,7 +989,13 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
         return backup_info
 
     def recover(
-        self, backup_info, dest, tablespaces=None, remote_command=None, **kwargs
+        self,
+        backup_info,
+        dest,
+        wal_dest=None,
+        tablespaces=None,
+        remote_command=None,
+        **kwargs
     ):
         """
         Performs a recovery of a backup
@@ -940,6 +1003,8 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
         :param barman.infofile.LocalBackupInfo backup_info: the backup
             to recover
         :param str dest: the destination directory
+        :param str|None wal_dest: the destination directory for WALs when doing PITR.
+            See :meth:`~barman.recovery_executor.RecoveryExecutor._set_pitr_targets` for more details.
         :param dict[str,str]|None tablespaces: a tablespace name -> location
             map (for relocation)
         :param str|None remote_command: default None. The remote command
@@ -987,6 +1052,7 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
             recovery_info = executor.recover(
                 backup_info,
                 dest,
+                wal_dest=wal_dest,
                 tablespaces=tablespaces,
                 remote_command=remote_command,
                 **kwargs
@@ -1138,22 +1204,56 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
             _logger.debug("Deleting PGDATA directory: %s" % pg_data)
             shutil.rmtree(pg_data)
 
+    def _run_pre_delete_wal_scripts(self, wal_info):
+        """
+        Run the pre-delete hook-scripts, if any, on the given WAL.
+
+        :param barman.infofile.WalFileInfo wal_info: WAL to run the script on.
+        """
+        # Run the pre_wal_delete_script if present.
+        script = HookScriptRunner(self, "wal_delete_script", "pre")
+        script.env_from_wal_info(wal_info)
+        script.run()
+        # Run the pre_wal_delete_retry_script if present.
+        retry_script = RetryHookScriptRunner(self, "wal_delete_retry_script", "pre")
+        retry_script.env_from_wal_info(wal_info)
+        retry_script.run()
+
+    def _run_post_delete_wal_scripts(self, wal_info, error=None):
+        """
+        Run the post-delete hook-scripts, if any, on the given WAL.
+
+        :param barman.infofile.WalFileInfo wal_info: WAL to run the script on.
+        :param None|str error: error message in case a failure happened.
+        """
+        # Run the post_wal_delete_retry_script if present.
+        try:
+            retry_script = RetryHookScriptRunner(
+                self, "wal_delete_retry_script", "post"
+            )
+            retry_script.env_from_wal_info(wal_info, None, error)
+            retry_script.run()
+        except AbortedRetryHookScript as e:
+            # Ignore the ABORT_STOP as it is a post-hook operation
+            _logger.warning(
+                "Ignoring stop request after receiving "
+                "abort (exit code %d) from post-wal-delete "
+                "retry hook script: %s",
+                e.hook.exit_status,
+                e.hook.script,
+            )
+        # Run the post_wal_delete_script if present.
+        script = HookScriptRunner(self, "wal_delete_script", "post")
+        script.env_from_wal_info(wal_info, None, error)
+        script.run()
+
     def delete_wal(self, wal_info):
         """
         Delete a WAL segment, with the given WalFileInfo
 
         :param barman.infofile.WalFileInfo wal_info: the WAL to delete
         """
-
-        # Run the pre_wal_delete_script if present.
-        script = HookScriptRunner(self, "wal_delete_script", "pre")
-        script.env_from_wal_info(wal_info)
-        script.run()
-
-        # Run the pre_wal_delete_retry_script if present.
-        retry_script = RetryHookScriptRunner(self, "wal_delete_retry_script", "pre")
-        retry_script.env_from_wal_info(wal_info)
-        retry_script.run()
+        self._run_pre_delete_wal_scripts(wal_info)
 
         error = None
         try:
@@ -1173,27 +1273,7 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
             )
             output.warning(error)
 
-        # Run the post_wal_delete_retry_script if present.
-        try:
-            retry_script = RetryHookScriptRunner(
-                self, "wal_delete_retry_script", "post"
-            )
-            retry_script.env_from_wal_info(wal_info, None, error)
-            retry_script.run()
-        except AbortedRetryHookScript as e:
-            # Ignore the ABORT_STOP as it is a post-hook operation
-            _logger.warning(
-                "Ignoring stop request after receiving "
-                "abort (exit code %d) from post-wal-delete "
-                "retry hook script: %s",
-                e.hook.exit_status,
-                e.hook.script,
-            )
-
-        # Run the post_wal_delete_script if present.
-        script = HookScriptRunner(self, "wal_delete_script", "post")
-        script.env_from_wal_info(wal_info, None, error)
-        script.run()
+        self._run_post_delete_wal_scripts(wal_info, error)
 
     def check(self, check_strategy):
         """
@@ -1314,82 +1394,6 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
         else:
             return {}
 
-    def rebuild_xlogdb(self):
-        """
-        Rebuild the whole xlog database guessing it from the archive content.
-        """
-        from os.path import isdir, join
-
-        output.info("Rebuilding xlogdb for server %s", self.config.name)
-        root = self.config.wals_directory
-        comp_manager = self.compression_manager
-        wal_count = label_count = history_count = 0
-        # lock the xlogdb as we are about replacing it completely
-        with self.server.xlogdb("w") as fxlogdb:
-            xlogdb_dir = os.path.dirname(fxlogdb.name)
-            with tempfile.TemporaryFile(mode="w+", dir=xlogdb_dir) as fxlogdb_new:
-                for name in sorted(os.listdir(root)):
-                    # ignore the xlogdb and its lockfile
-                    if name.startswith(self.server.XLOG_DB):
-                        continue
-                    fullname = join(root, name)
-                    if isdir(fullname):
-                        # all relevant files are in subdirectories
-                        hash_dir = fullname
-                        for wal_name in sorted(os.listdir(hash_dir)):
-                            fullname = join(hash_dir, wal_name)
-                            if isdir(fullname):
-                                _logger.warning(
-                                    "unexpected directory "
-                                    "rebuilding the wal database: %s",
-                                    fullname,
-                                )
-                            else:
-                                if xlog.is_wal_file(fullname):
-                                    wal_count += 1
-                                elif xlog.is_backup_file(fullname):
-                                    label_count += 1
-                                elif fullname.endswith(".tmp"):
-                                    _logger.warning(
-                                        "temporary file found "
-                                        "rebuilding the wal database: %s",
-                                        fullname,
-                                    )
-                                    continue
-                                else:
-                                    _logger.warning(
-                                        "unexpected file "
-                                        "rebuilding the wal database: %s",
-                                        fullname,
-                                    )
-                                    continue
-                                wal_info = comp_manager.get_wal_file_info(fullname)
-                                fxlogdb_new.write(wal_info.to_xlogdb_line())
-                    else:
-                        # only history files are here
-                        if xlog.is_history_file(fullname):
-                            history_count += 1
-                            wal_info = comp_manager.get_wal_file_info(fullname)
-                            fxlogdb_new.write(wal_info.to_xlogdb_line())
-                        else:
-                            _logger.warning(
-                                "unexpected file rebuilding the wal database: %s",
-                                fullname,
-                            )
-                fxlogdb_new.flush()
-                fxlogdb_new.seek(0)
-                fxlogdb.seek(0)
-                shutil.copyfileobj(fxlogdb_new, fxlogdb)
-                fxlogdb.truncate()
-        output.info(
-            "Done rebuilding xlogdb for server %s "
-            "(history: %s, backup_labels: %s, wal_file: %s)",
-            self.config.name,
-            history_count,
-            label_count,
-            wal_count,
-        )
-
     def get_latest_archived_wals_info(self):
         """
         Return a dictionary of timelines associated with the
@@ -1457,7 +1461,9 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
             tuples which define inclusive ranges of WALs which must not be deleted.
         :return list: a list of removed WAL files
         """
-        removed = []
+        # A dictionary where key is the WAL directory name and value is a list of
+        # wal_info object representing the WALs to be deleted in that directory
+        wals_to_remove = defaultdict(list)
         with self.server.xlogdb("r+") as fxlogdb:
             xlogdb_dir = os.path.dirname(fxlogdb.name)
             with tempfile.TemporaryFile(mode="w+", dir=xlogdb_dir) as fxlogdb_new:
@@ -1498,18 +1504,59 @@ class BackupManager(RemoteStatusMixin, KeepManagerMixin):
                         keep |= wal_info.name >= backup_info.begin_wal
 
                     # If the file has to be kept write it in the new xlogdb
-                    # otherwise delete it  and record it in the removed list
+                    # otherwise add it to the removal list
                     if keep:
                         fxlogdb_new.write(wal_info.to_xlogdb_line())
                     else:
-                        self.delete_wal(wal_info)
-                        removed.append(wal_info.name)
+                        wal_dir = os.path.dirname(wal_info.fullpath(self.server))
+                        wals_to_remove[wal_dir].append(wal_info)
+
+                wals_removed = self.delete_wals(wals_to_remove)
+
                 fxlogdb_new.flush()
                 fxlogdb_new.seek(0)
                 fxlogdb.seek(0)
                 shutil.copyfileobj(fxlogdb_new, fxlogdb)
                 fxlogdb.truncate()
-        return removed
+
+        return wals_removed
+
+    def delete_wals(self, wals_to_delete):
+        """
+        Delete the given WAL files. The entire WAL directory is deleted when possible.
+
+        :param dict[str, list[WalFileInfo]] wals_to_delete: A dictionary where key is the WAL directory name and
+            value is a list of wal_info objects representing the WALs to be deleted in
+            that directory.
+        :return list[str]: a list of deleted WAL names.
+        """
+        wals_deleted = []
+        for wal_dir, wal_list in wals_to_delete.items():
+            delete_directory = False
+            # Each directory can contain up to 256 WAL files. If the deletion list
+            # contains 256 entries, the entire directory can be safely deleted
+            # Otherwise, check if all WALs in the directory are in the deletion list
+            if len(wal_list) >= 256:
+                delete_directory = True
+            else:
+                wal_names_to_delete = {wal_info.name for wal_info in wal_list}
+                wal_names_in_dir = os.listdir(wal_dir)
+                if set(wal_names_in_dir).issubset(wal_names_to_delete):
+                    delete_directory = True
+            # If the directory can be deleted, run the hook-scripts on each WAL file
+            # before and after the rmtree. Otherwise, delete each WAL individually
+            if delete_directory:
+                for wal_info in wal_list:
+                    self._run_pre_delete_wal_scripts(wal_info)
+                shutil.rmtree(wal_dir)
+                for wal_info in wal_list:
+                    self._run_post_delete_wal_scripts(wal_info)
+                    wals_deleted.append(wal_info.name)
+            else:
+                for wal_info in wal_list:
+                    self.delete_wal(wal_info)
+                    wals_deleted.append(wal_info.name)
+        return wals_deleted
 
     def validate_last_backup_maximum_age(self, last_backup_maximum_age):
         """
